import numpy as np
from dataclasses import dataclass, asdict

#: Default number of heights
N_H = 8
#: Default number of segments
N_D = 600
#: Default number of speeds
N_S = 7
#: Default min speed (in *mach*)
S_MIN = 0.6
#: Default max speed (in *mach*)
S_MAX = 0.9
#: Default segment length (in *km*)
WAYPOINT = 10.0
#: Default Mach speed (in *km/s*)
MAC = 0.3403
#: Default speed cost (additional speed cost is :math:`S * s^2`)
SPEED_COST = 50.0
#: Default base cost
CRUISE_COST = 20.0
#: Default climbing cost
CLIMB_COST = 100.0
#: Default climbing (down) gain
DOWN_GAIN = 25.0
#: Default cost reduction due to height
HEIGHT_GAIN = 3.0
#: Default weight that increases cost by 1
INVERSE_WEIGHT = 4000.0


@dataclass
class CostModel:
    """
    Energy consumption and time cost(s) depending on various parameters.
    All values are expressed in units per waypoint.

    The model is generic and applies to any vehicle with state-dependent energy costs.
    "Height" represents any discrete state dimension (altitude, depth, elevation, etc.),
    and "weight" represents remaining energy (fuel, battery charge, etc.).

    Attributes
    ----------

    timings: :class:`~numpy.ndarray`
        Time to travel the waypoint w.r.t. speed index.
    speed_array: :class:`~numpy.ndarray`
        Energy cost to travel the waypoint w.r.t. speed index.
    cruise_matrix: :class:`~numpy.ndarray`
        Base cost w.r.t. waypoint and height indices.
    climb_matrix: :class:`~numpy.ndarray`
        Cost to increase height w.r.t. waypoint and height indices.
    down_matrix: :class:`~numpy.ndarray`
        Gain from decreasing height w.r.t. waypoint and height indices.
    height_gain: :class:`float`
        Cost reduction w.r.t. height index.
    weight_cost: :class:`float`
        Cost modifier w.r.t. remaining energy (spare weight).
    """

    timings: np.ndarray
    speed_array: np.ndarray
    cruise_matrix: np.ndarray
    climb_matrix: np.ndarray
    down_matrix: np.ndarray
    height_gain: float
    weight_cost: float

    @property
    def dict(self):
        return asdict(self)


class CostRandom(CostModel):
    """
    A simple cost model with global parameters and randomized terrain.

    Default parameters are tuned for aircraft cruise, but the model
    generalizes to any vehicle. Adapt parameters for your domain.

    Parameters
    ----------
    n_h: :class:`int`, optional
        Number of height/state steps.
    n_d: :class:`int`, optional
        Number of waypoints.
    n_s: :class:`int`, optional
        Number of speed levels.
    s_min: :class:`float`, optional
        Minimum speed (default: Mach 0.6 for aircraft).
    s_max: :class:`float`, optional
        Maximum speed (default: Mach 0.9 for aircraft).
    waypoint: :class:`float`, optional
        Waypoint length (in km).
    mac: :class:`float`, optional
        Reference speed (default: Mach 1 in km/s for aircraft).
    speed_cost: :class:`float`, optional
        Energy cost coefficient for speed.
    cruise_cost: :class:`float`, optional
        Base energy cost per waypoint.
    climb_cost: :class:`float`, optional
        Energy cost for increasing height/state.
    down_gain: :class:`float`, optional
        Energy recovered when decreasing height/state.
    height_gain: :class:`float`, optional
        Efficiency improvement from higher states.
    inverse_weight: :class:`float`, optional
        Energy level at which cost increases by 1 unit.
    seed: :class:`int`, optional
        Random seed for reproducible terrain generation.
    """

    def __init__(
        self,
        n_h=N_H,
        n_d=N_D,
        n_s=N_S,
        s_min=S_MIN,
        s_max=S_MAX,
        waypoint=WAYPOINT,
        mac=MAC,
        speed_cost=SPEED_COST,
        cruise_cost=CRUISE_COST,
        climb_cost=CLIMB_COST,
        down_gain=DOWN_GAIN,
        height_gain=HEIGHT_GAIN,
        inverse_weight=INVERSE_WEIGHT,
        seed=None,
    ):
        if seed is not None:
            np.random.seed(seed)
        speeds = np.linspace(s_min, s_max, n_s)
        timings = waypoint / mac / speeds
        speed_array = speed_cost * speeds**2
        cruise_matrix = cruise_cost * (0.5 + np.random.rand(n_d, n_h))
        climb_matrix = climb_cost * (0.5 + np.random.rand(n_d, n_h))
        down_matrix = down_gain * (0.5 + np.random.rand(n_d, n_h))
        super().__init__(
            timings=timings,
            speed_array=speed_array,
            cruise_matrix=cruise_matrix,
            climb_matrix=climb_matrix,
            down_matrix=down_matrix,
            height_gain=height_gain,
            weight_cost=1 / inverse_weight,
        )



import dill as pickle
import zstandard as zstd
import errno
import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from contextlib import contextmanager


@contextmanager
def safe_write(path):
    """
    Context manager to write a file in two steps: first use a tmp file,
    then rename the file if everything went well.
    In case of error, the temp file is deleted.
    """
    path = Path(path)
    with NamedTemporaryFile(mode="wb", dir=path.parent, delete=False) as tmpfile:
        tmp_path = Path(tmpfile.name)
        try:
            yield tmpfile
            # Proper closure of temp file
            tmpfile.close()
            # Atomic renaming
            tmp_path.replace(path)
        except Exception:
            # Error => delete the temp file
            if tmp_path.exists():
                tmp_path.unlink()
            raise  # Propagate error


class MixInIO:
    """
    Provide basic save/load capacities to other classes.
    """

    def dump(
        self, filename: str, path=".", overwrite=False, compress=True, stemize=True
    ):
        """
        Save instance to file.

        Parameters
        ----------
        filename: str
            The stem of the filename.
        path: :py:class:`str` or :py:class:`~pathlib.Path`, optional
            The location path.
        overwrite: bool, default=False
            Should existing file be erased if it exists?
        compress: bool, default=True
            Should Zstd compression be used?
        stemize: bool, default=True
            Trim any extension (e.g. .xxx)

        Examples
        ----------

        >>> import tempfile
        >>> v1 = ToyClass(42)
        >>> v2 = ToyClass()
        >>> v2.value
        0
        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.dump(filename='myfile', compress=True, path=tmpdirname)
        ...     dir_content = [file.name for file in Path(tmpdirname).glob('*')]
        ...     v2 = ToyClass.load(filename='myfile', path=Path(tmpdirname))
        ...     v1.dump(filename='myfile', compress=True, path=tmpdirname) # doctest.ELLIPSIS
        File ...myfile.pkl.zst already exists! Use overwrite option to overwrite.
        >>> dir_content
        ['myfile.pkl.zst']
        >>> v2.value
        42

        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.dump(filename='myfile', compress=False, path=tmpdirname)
        ...     v1.dump(filename='myfile', compress=False, path=tmpdirname) # doctest.ELLIPSIS
        File ...myfile.pkl already exists! Use overwrite option to overwrite.

        >>> v1.value = 51
        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.dump(filename='myfile', path=tmpdirname, compress=False)
        ...     v1.dump(filename='myfile', path=tmpdirname, overwrite=True, compress=False)
        ...     v2 = ToyClass.load(filename='myfile', path=tmpdirname)
        ...     dir_content = [file.name for file in Path(tmpdirname).glob('*')]
        >>> dir_content
        ['myfile.pkl']
        >>> v2.value
        51

        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...    v2 = ToyClass.load(filename='thisfilenamedoesnotexist')
        Traceback (most recent call last):
         ...
        FileNotFoundError: [Errno 2] No such file or directory: ...
        """
        path = Path(path)
        fn = Path(filename)
        if stemize:
            fn = Path(fn.stem)
        if compress:
            destination = path / (fn.name + ".pkl.zst")
            if destination.exists() and not overwrite:
                print(
                    f"File {destination} already exists! Use overwrite option to overwrite."
                )
            else:
                with safe_write(destination) as f:
                    cctx = zstd.ZstdCompressor(level=3)
                    with cctx.stream_writer(f) as z:
                        pickle.dump(self, z, protocol=5)
        else:
            destination = path / (fn.name + ".pkl")
            if destination.exists() and not overwrite:
                print(
                    f"File {destination} already exists! Use overwrite option to overwrite."
                )
            else:
                with safe_write(destination) as f:
                    pickle.dump(self, f)

    @classmethod
    def load(cls, filename: str, path="."):
        """
        Load instance from file.

        Parameters
        ----------
        filename: str
            The stem of the filename.
        path: :py:class:`str` or :py:class:`~pathlib.Path`, optional
            The location path.
        """
        path = Path(path)
        dest = path / Path(filename).with_suffix(".pkl")
        if dest.exists():
            with open(dest, "rb") as f:
                return pickle.load(f)
        else:
            dest = dest.with_suffix(".pkl.zst")
            if dest.exists():
                dctx = zstd.ZstdDecompressor()
                # Load compressed data
                with open(dest, "rb") as f, dctx.stream_reader(f) as z:
                    return pickle.load(z)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), dest)


class ToyClass(MixInIO):
    def __init__(self, value=0):
        self.value = value



"""Top-level package for Smart Cruise."""

from importlib.metadata import metadata
from smart_cruise.models import CostModel as CostModel
from smart_cruise.models import CostRandom as CostRandom
from smart_cruise.smart_cruise import Cruise as Cruise

infos = metadata(__name__)
__version__ = infos["Version"]
__author__ = "Fabien Mathieu"
__email__ = "fabien.mathieu@normalesup.org"



from dataclasses import dataclass

import numpy as np
from matplotlib import pyplot as plt

from smart_cruise.jit import compute

#: Default weight budget.
W0 = 26000.0

#: Default time budget.
T0 = 26000.0

#: Default backoff for changing speed.
BACKOFF = 9

#: Default discretization of the Pareto front of individual states.
PARETO_MAX = 30


class Trajectories:
    """
    Handler class for optimal trajectories.

    Parameters
    ----------
    trajs: :class:`~numpy.ndarray`
        :math:`n_t \\times n_d` array of :math:`(h, s, w, t)` states.
    n_h: :class:`int`
        Number of heights (for normalization).
    n_s: :class:`int`
        Number of speeds (for normalization).
    """

    def __init__(self, trajs, n_h, n_s):
        self.trajs = trajs
        self.n_h = n_h
        self.n_s = n_s

    def get_traj(self, i):
        """
        Get trajectory profile.

        Parameters
        ----------
        i: :class:`int`
            Index of trajectory.

        Returns
        -------
        h: :class:`~numpy.ndarray`
        s: :class:`~numpy.ndarray`
        w: :class:`~numpy.ndarray`
        t: :class:`~numpy.ndarray`
        """
        h = np.array([t[0] for t in self.trajs[i, :]]) / self.n_h
        s = np.array([t[1] for t in self.trajs[i, :]]) / self.n_s
        w = np.array([t[2] for t in self.trajs[i, :]]) / self.trajs[0, 0][2]
        t = np.array([t[3] for t in self.trajs[i, :]]) / self.trajs[0, 0][3]
        return h, s, w, t

    def plot_traj(self, i):
        """
        Display trajectory profile.

        Parameters
        ----------
        i: :class:`int`
            Index of trajectory.

        Returns
        -------
        None
        """
        h, s, w, t = self.get_traj(i)
        plt.plot(h, label="Height")
        plt.plot(s, label="Speed")
        plt.plot(w, label="Remaining Energy")
        plt.plot(t, label="Spare Time")
        plt.ylim([0, 1])
        plt.xlim([0, len(h) - 1])
        plt.legend()
        plt.show()

    def get_front(self):
        """
        Get Pareto front.

        Returns
        -------
        w: :class:`~numpy.ndarray`
        t: :class:`~numpy.ndarray`
        """
        w = np.array([t[2] for t in self.trajs[:, -1]], dtype=float)
        t = np.array([t[3] for t in self.trajs[:, -1]], dtype=float)
        return w, t

    def plot_front(self):
        """
        Display Pareto front.

        Returns
        -------
        None
        """
        w, t = self.get_front()
        plt.plot(w, t)
        plt.xlabel("Remaining energy")
        plt.ylabel("Spare time")
        plt.xlim([0, None])
        plt.ylim([0, None])
        plt.show()


@dataclass
class CruiseParameters:
    """
    Attributes
    ----------

    backoff: :class:`int`, default=BACKOFF
        Timer to change speed.
    pareto_max: :class:`int`, default=PARETO_MAX
        Max number of pareto-optimal points per state.
    """

    backoff: int = BACKOFF
    pareto_max: int = PARETO_MAX


class Cruise:
    """

    Parameters
    ----------
    model: :class:`~smart_cruise.models.CostModel`
        The cost model.
    parameters: :class:`~smart_cruise.smart_cruise.CruiseParameters`
        Simulation parameters.

    Examples
    --------

    >>> from smart_cruise import CostRandom
    >>> model = CostRandom(n_d=100, seed=42)
    >>> cruise = Cruise(model)
    >>> cruise.parameters.pareto_max = 5
    >>> cruise.compute()
    >>> w, t = cruise.trajectories.get_front()

    Weight optimal values:

    >>> f"Weight: {w[0]:.2f}, time: {t[0]:.2f}"
    'Weight: 23160.79, time: 21102.36'

    Time optimal values:
    >>> f"Weight: {w[-1]:.2f}, time: {t[-1]:.2f}"
    'Weight: 20938.41, time: 22734.91'
    """

    def __init__(self, model=None, parameters=None):
        self.model = model
        if parameters is None:
            parameters = CruiseParameters()
        self.parameters = parameters
        self.trajectories = None

    def compute(self, w0=W0, t0=T0):
        trajectories = compute(
            [(w0, t0)],
            backoff=self.parameters.backoff,
            pareto_max=self.parameters.pareto_max,
            **self.model.dict,
        )
        self.trajectories = Trajectories(
            trajectories,
            n_s=len(self.model.timings),
            n_h=self.model.cruise_matrix.shape[1],
        )



import numpy as np
from numba import njit
from numba.typed import List
from numba.core import config as numba_config


@njit
def new_state_dict():
    """
    Creates empty state dict

    Returns
    -------
    :class:`~numba.typed.Dict`
        A Numba typed dict (h, s, b) -> [(w, t, p)].
    """
    return {(1, 2, 3): List([(1.0, 2.0, 3)]) for _ in range(0)}


@njit
def insert_state(states, h, s, b, w, t, p):
    """
    Defaultdict-like state insertion.

    Parameters
    ----------
    states: :class:`~numba.typed.Dict`
        A Numba typed dict (h, s, b) -> [(w, t, p)].
    h: :class:`int`
        Height index.
    s: :class:`int`
        Speed index.
    b: :class:`int`
        Backoff.
    w: :class:`float`
        Weight.
    t: :class:`float`
        Time.
    p: :class:`int`
        Index of origin state.

    Returns
    -------
    None
    """
    if w > 0 and t > 0:
        if (h, s, b) in states:
            states[h, s, b].append((w, t, p))
        else:
            states[h, s, b] = List([(w, t, p)])


np_rec = np.dtype([("h", "i8"), ("s", "i8"), ("w", "f8"), ("t", "f8")])

if numba_config.DISABLE_JIT:
    # Pure Python mode: use numpy dtypes directly
    wt_type = np.dtype([("f0", "f8"), ("f1", "f8")])
    state_type = np.dtype(
        [("h", "i8"), ("s", "i8"), ("b", "i8"), ("w", "f8"), ("t", "f8"), ("p", "i8")]
    )
    nb_rec = np_rec
else:
    # JIT mode: use Numba types
    from numba import typeof, from_dtype

    wt_type = typeof((1.0, 2.0))
    state_type = typeof((1, 2, 3, 4.0, 5.0, 6))
    nb_rec = from_dtype(np_rec)


@njit
def pareto_collapse(states, pareto_max):
    """
    Parameters
    ----------
    states: :class:`~numba.typed.Dict`
        A Numba typed dict (h, s, b) -> [(w, t, p)].
    pareto_max: :class:`int`
        Maximum number of pareto points.

    Returns
    -------
    None
    """
    for state, wtp in states.items():
        keys = np.empty(len(wtp), dtype=wt_type)
        for i, x in enumerate(wtp):
            keys[i] = (-x[0], -x[1])
        indices = np.argsort(keys)
        wtp0 = wtp[indices[0]]
        res = List([wtp0])
        tm = wtp0[1]
        for i in indices[1:]:
            t = wtp[i][1]
            if t > tm:
                res.append(wtp[i])
                tm = t

        # Sample if needed
        if len(res) > pareto_max:
            res = List(
                [res[round(y)] for y in np.linspace(0, len(res) - 1, pareto_max)]
            )
        states[state] = res


@njit
def compute(
    start,
    timings,
    speed_array,
    cruise_matrix,
    climb_matrix,
    down_matrix,
    height_gain,
    weight_cost,
    backoff,
    pareto_max,
):
    """
    Core computation of optimal trajectories.

    Parameters
    ----------
    start: :class:`list`
        Initial (weight, time) budgets.
    timings: :class:`~numpy.ndarray`
        Time cost per speed level.
    speed_array: :class:`~numpy.ndarray`
        Speed cost adjustments.
    cruise_matrix: :class:`~numpy.ndarray`
        Base energy cost matrix (waypoint x height).
    climb_matrix: :class:`~numpy.ndarray`
        Energy cost for climbing one height level.
    down_matrix: :class:`~numpy.ndarray`
        Energy gain for descending one height level.
    height_gain: :class:`float`
    weight_cost: :class:`float`
    backoff: :class:`int`
    pareto_max: :class:`int`
    """
    waypoint = np.empty(len(start), dtype=state_type)
    for i, wt in enumerate(start):
        waypoint[i] = (0, 0, 0, *wt, 0)
    waypoints = [waypoint]
    n_d, n_h = cruise_matrix.shape
    n_s = len(speed_array)
    # One waypoint after the other
    for i in range(n_d):
        # new state dict
        states = new_state_dict()
        # Browse states of current waypoint and processes their successors
        for p, state in enumerate(waypoints[i]):
            h, s, b, w, t, _ = state
            base_cost = cruise_matrix[i, h] - h * height_gain + w * weight_cost
            c = climb_matrix[i, h]
            d = down_matrix[i, h]
            bb = max(0, b - 1)
            ww = w - max(0.0, base_cost + speed_array[s])
            tt = t - timings[s]
            insert_state(states, h, s, bb, ww, tt, p)
            if h < n_h - 1:
                insert_state(states, h + 1, s, bb, ww - c, tt, p)
            if h > 0:
                insert_state(states, h - 1, s, bb, ww + d, tt, p)
            if b == 0:
                for ss in range(n_s):
                    if ss != s:
                        ww = w - max(0.0, base_cost + speed_array[ss])
                        tt = t - timings[ss]
                        insert_state(states, h, ss, backoff, ww, tt, p)
                        if h < n_h - 1:
                            insert_state(states, h + 1, ss, backoff, ww - c, tt, p)
                        if h > 0:
                            insert_state(states, h - 1, ss, backoff, ww + d, tt, p)
        pareto_collapse(states, pareto_max=pareto_max)
        n_states = sum([len(v) for v in states.values()])
        waypoint = np.empty(n_states, dtype=state_type)
        index = 0
        for k, vv in states.items():
            for v in vv:
                waypoint[index] = (*k, *v)
                index += 1
        waypoints.append(waypoint)

    # Global Pareto-reduction of last waypoint
    last = waypoints[-1]
    if len(last) == 0:
        return None
    keys = np.empty(len(last), dtype=wt_type)
    for i, x in enumerate(last):
        keys[i] = (-x[3], -x[4])
    indices = np.argsort(keys)
    ends = [last[indices[0]]]
    tm = last[indices[0]][4]
    for i in indices[1:]:
        state = last[i]
        if state[4] > tm:
            ends.append(state)
            tm = state[4]

    # Convert to trajectory array
    n_traj = len(ends)
    trajectories = np.empty((n_traj, n_d + 1), dtype=nb_rec)
    for i, end in enumerate(ends):
        trajectories[i, -1]["h"] = end[0]
        trajectories[i, -1]["s"] = end[1]
        trajectories[i, -1]["w"] = end[3]
        trajectories[i, -1]["t"] = end[4]
        pos = end[5]
        for j in range(n_d - 1, -1, -1):
            state = waypoints[j][pos]
            trajectories[i, j]["h"] = state[0]
            trajectories[i, j]["s"] = state[1]
            trajectories[i, j]["w"] = state[3]
            trajectories[i, j]["t"] = state[4]
            pos = state[5]

    return trajectories
